package edu.illinois.reex.instrumentation;

import java.util.Set;

import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Opcodes;
import org.objectweb.asm.Type;

/**
 * ReEx instrumentor for shared field read/writes.
 * 
 * @author Vilas Jagannath <vbangal2@illinois.edu>
 * 
 */
public class SharedAccessEventsMethodTransformer extends LocationAwareLocalVariablesSorter {

    private static final String PERFORM_UNLOCK = "performUnlock";
    private static final String PERFORM_LOCK = "performLock";
    private static final String AFTER_FIELD_ACCESS = "afterFieldAccess";
    private static final String BEFORE_FIELD_ACCESS = "beforeFieldAccess";
    private static final String OBJECT_VOID = "(Ljava/lang/Object;)V";
    private static final String BOOL_3STRINGS_VOID = "(ZLjava/lang/String;Ljava/lang/String;Ljava/lang/String;)V";

    private final boolean insertSyncBlock;
    private final boolean staticSyncBlock;

    private int lockVariable;
    private final Set<String> nonArrayfinalFields;

    public SharedAccessEventsMethodTransformer(int access, String desc, String methodName, String className, MethodVisitor methodVisitor,
            boolean insertSyncBlock, boolean staticSyncBlock, Set<String> nonArrayFinalFields) {
        super(access, desc, methodName, className, methodVisitor);
        this.insertSyncBlock = insertSyncBlock;
        this.staticSyncBlock = staticSyncBlock;
        this.nonArrayfinalFields = nonArrayFinalFields;
    }

    @Override
    protected void onMethodEnter() {
        if (this.insertSyncBlock) {
            Type classType = Type.getType("L" + this.className + ";");
            if (this.staticSyncBlock) {
                super.mv.visitLdcInsn(classType);
                this.lockVariable = newLocal(classType);
            } else {
                super.mv.visitVarInsn(Opcodes.ALOAD, 0);
                this.lockVariable = newLocal(classType);
            }
            super.mv.visitInsn(Opcodes.DUP);
            super.mv.visitVarInsn(Opcodes.ASTORE, this.lockVariable);
            visitInsn(Opcodes.MONITORENTER);
        }
    }

    @Override
    protected void onMethodExit(int opcode) {
        if (this.insertSyncBlock) {
            super.mv.visitVarInsn(Opcodes.ALOAD, this.lockVariable);
            visitInsn(Opcodes.MONITOREXIT);
        }
    }

    @Override
    public void visitFieldInsn(int opcode, String owner, String name, String desc) {
        /*
         * Final field access is not a scheduling event (only detecting local
         * (same class) non array final fields atm). System.{err,out}.* is not
         * schedule relevant either.
         */
        if ((this.nonArrayfinalFields.contains(name) && owner.equals(this.className)) || owner.equals("java/lang/System")) {
            super.visitFieldInsn(opcode, owner, name, desc);
        } else {
            boolean isRead = false;
            if (opcode == Opcodes.GETFIELD || opcode == Opcodes.GETSTATIC) {
                isRead = true;
            }
            super.updateThreadLocation();
            this.informSchedulerAboutFieldAccess(true, isRead, owner, name, desc);
            super.visitFieldInsn(opcode, owner, name, desc);
            this.informSchedulerAboutFieldAccess(false, isRead, owner, name, desc);
        }
    }

    @Override
    public void visitInsn(int opcode) {
        if (opcode == Opcodes.MONITORENTER) {
            this.performLockEventUsingScheduler(true);
        } else if (opcode == Opcodes.MONITOREXIT) {
            this.performLockEventUsingScheduler(false);
        } else {
            super.visitInsn(opcode);
        }
    }

    private void informSchedulerAboutFieldAccess(boolean isBefore, boolean isRead, String fieldOwner, String fieldName, String fieldDesc) {
        super.mv.visitInsn(isRead ? Opcodes.ICONST_1 : Opcodes.ICONST_0);
        super.mv.visitLdcInsn(fieldOwner);
        super.mv.visitLdcInsn(fieldName);
        super.mv.visitLdcInsn(fieldDesc);
        if (isBefore) {
            super.mv.visitMethodInsn(Opcodes.INVOKESTATIC, Instrumentor.INSTR_EVENTS_RECEIVER, BEFORE_FIELD_ACCESS, BOOL_3STRINGS_VOID);
        } else {
            super.mv.visitMethodInsn(Opcodes.INVOKESTATIC, Instrumentor.INSTR_EVENTS_RECEIVER, AFTER_FIELD_ACCESS, BOOL_3STRINGS_VOID);
        }
    }

    private void performLockEventUsingScheduler(boolean isLock) {
        super.updateThreadLocation();
        if (isLock) {
            super.mv.visitMethodInsn(Opcodes.INVOKESTATIC, Instrumentor.INSTR_EVENTS_RECEIVER, PERFORM_LOCK, OBJECT_VOID);
        } else {
            super.mv.visitMethodInsn(Opcodes.INVOKESTATIC, Instrumentor.INSTR_EVENTS_RECEIVER, PERFORM_UNLOCK, OBJECT_VOID);
        }
    }

}
