package edu.illinois.reex.listeners;

import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.SortedSet;

import edu.illinois.reex.ReExProperties;
import edu.illinois.reex.scheduling.ChoiceType;
import edu.illinois.reex.scheduling.ThreadInfo;
import edu.illinois.reex.scheduling.events.EventDesc;

/**
 * 
 * @author Vilas Jagannath <vbangal2@illinois.edu>
 * 
 */
public class Listeners {

    private static final String DEBUG_EXPLORATION_DEFAULT = "false";

    private static final Set<ExplorationListener> explorationListeners;

    public static final boolean debugExploration;

    static {
        ReExProperties reexProps = ReExProperties.getInstance();
        explorationListeners = new HashSet<ExplorationListener>();

        if (debugExploration = Boolean.parseBoolean(reexProps.getProperty(ReExProperties.EXPLORATION_DEBUG_KEY, DEBUG_EXPLORATION_DEFAULT))) {
            explorationListeners.add(new ExplorationDebugListener());
        }

        /* Add the other listeners to be used */
        String listenerClassNames = reexProps.getProperty(ReExProperties.LISTENERS_KEY, ExplorationStatsListener.class.getName());
        if (listenerClassNames != null) {
            for (String listenerClassName : listenerClassNames.split(",")) {
                try {
                    ExplorationListener listener = (ExplorationListener) Class.forName(listenerClassName).newInstance();
                    explorationListeners.add(listener);
                } catch (Exception e) {
                    System.err.println("Unable to obtain instance of: " + listenerClassName);
                    e.printStackTrace();
                }
            }
        }
    }

    public static void addListener(ExplorationListener listener) {
        explorationListeners.add(listener);
    }

    public static void removeListener(ExplorationListener listener) {
        explorationListeners.remove(listener);
    }

    public static void fireStartingExploration(String name) {
        for (ExplorationListener listener : explorationListeners) {
            listener.startingExploration(name);
        }
    }

    public static void fireStartingSchedule() {
        for (ExplorationListener listener : explorationListeners) {
            listener.startingSchedule();
        }
    }

    public static void fireBeforeForking(ThreadInfo childThread) {
        for (ExplorationListener listener : explorationListeners) {
            listener.beforeForking(childThread);
        }
    }

    public static void fireBeforeEvent(EventDesc eventDesc) {
        for (ExplorationListener listener : explorationListeners) {
            listener.beforeEvent(eventDesc);
        }
    }

    public static void fireAfterEvent(EventDesc eventDesc) {
        for (ExplorationListener listener : explorationListeners) {
            listener.afterEvent(eventDesc);
        }
    }

    public static void fireMakingChoice(SortedSet<? extends Object> objectChoices, ChoiceType choiceType) {
        for (ExplorationListener listener : explorationListeners) {
            listener.makingChoice(objectChoices, choiceType);
        }
    }

    public static void fireChoiceMade(Object choice) {
        for (ExplorationListener listener : explorationListeners) {
            listener.choiceMade(choice);
        }
    }

    public static void fireCompletedSchedule(List<Integer> choicesMade) {
        for (ExplorationListener listener : explorationListeners) {
            listener.completedSchedule(choicesMade);
        }
    }

    public static void fireCompletedExploration() {
        for (ExplorationListener listener : explorationListeners) {
            listener.completedExploration();
        }
    }

    public static void fireFailureDetected(String errorMsg, List<Integer> choicesMade) {
        for (ExplorationListener listener : explorationListeners) {
            listener.failureDetected(errorMsg, choicesMade);
        }
    }

}
