package edu.illinois.reex.scheduling;

import java.lang.Thread.UncaughtExceptionHandler;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.SortedMap;
import java.util.SortedSet;
import java.util.TreeMap;
import java.util.TreeSet;

import sun.misc.Unsafe;
import edu.illinois.reex.ReExProperties;
import edu.illinois.reex.internaljuc.locks.Reex_Condition;
import edu.illinois.reex.internaljuc.locks.Reex_ReentrantLock;
import edu.illinois.reex.listeners.Listeners;
import edu.illinois.reex.scheduling.events.EventDesc;
import edu.illinois.reex.scheduling.events.EventType;
import edu.illinois.reex.scheduling.events.FieldAccessEventDesc;
import edu.illinois.reex.scheduling.events.JoinEventDesc;
import edu.illinois.reex.scheduling.events.LocationDesc;
import edu.illinois.reex.scheduling.events.LockEventDesc;
import edu.illinois.reex.scheduling.events.ParkUnparkEventDesc;
import edu.illinois.reex.scheduling.events.ThreadLifeEventDesc;
import edu.illinois.reex.scheduling.events.WaitNotifyEventDesc;
import edu.illinois.reex.scheduling.filtering.DefaultFilter;
import edu.illinois.reex.scheduling.filtering.SchedulingFilter;
import edu.illinois.reex.scheduling.strategy.DefaultStrategy;
import edu.illinois.reex.scheduling.strategy.SchedulingStrategy;

/**
 * Class that orchestrates threads so that they can be scheduled using custom
 * {@link SchedulingStrategy}s. The instrumentation hooks all call into this
 * class.
 * 
 * @author Vilas Jagannath <vbangal2@illinois.edu>
 * @author Matt Kirn <kirn1@illinois.edu>
 * 
 */
public class Scheduler {

    private static final String BANG = "!";
    private static final String UNABLE_TO_OBTAIN_INSTANCE_OF = "Unable to obtain instance of: ";

    /******************************************************************
     ****************************************************************** 
     *********************** SCHEDULER STATE **************************
     ****************************************************************** 
     ******************************************************************/

    private static Map<Thread, ThreadInfo> liveThreadInfos;
    private static SortedSet<ThreadInfo> pausedThreadInfos;
    private static Set<ThreadInfo> blockedThreadInfos;
    private static final Reex_ReentrantLock schedulerStateLock = new Reex_ReentrantLock();
    private static final Reex_Condition schedulerWakeupCondition = schedulerStateLock.newCondition();

    /**
     * {@link SchedulingStrategy} to be used for scheduling decisions.
     */
    private static SchedulingStrategy schedulingStrategy;

    /**
     * {@link SchedulingFilter} to be used for scheduling decisions.
     */
    private static SchedulingFilter schedulingFilter;
    private static boolean exploring = false;

    private static final Unsafe unsafe = Unsafe.getUnsafe();

    /**
     * Initialize state before everything.
     */
    static {
        reinitState();

        // Catch any uncaught exception thrown by any thread
        // NOTE: this can be overridden by the code under test
        Thread.setDefaultUncaughtExceptionHandler(new UncaughtExceptionHandler() {

            /*
             * This method will execute in the context of the thread that raised
             * the exception
             */
            @Override
            public void uncaughtException(Thread t, Throwable e) {
                if (e != null) {
                    e.printStackTrace();
                }
                failureDetected(null);
                Listeners.fireCompletedExploration();
                System.exit(2);
            }
        });

        ReExProperties reexProps = ReExProperties.getInstance();
        /* Set the scheduling strategy to be used */
        String schedulingStrategyClassName = reexProps.getProperty(ReExProperties.SCHEDULING_STRATEGY_KEY);
        if (schedulingStrategyClassName != null) {
            try {
                schedulingStrategy = (SchedulingStrategy) Class.forName(schedulingStrategyClassName).newInstance();
                System.out.println("\nUsing the following scheduling strategy for exploration: " + schedulingStrategy.getClass().getName());
            } catch (Exception e) {
                System.err.println(UNABLE_TO_OBTAIN_INSTANCE_OF + schedulingStrategyClassName + BANG);
                e.printStackTrace();
                System.exit(2);
            }
        } else {
            System.out.println("\nWARNING: no value specified for property \"" + ReExProperties.SCHEDULING_STRATEGY_KEY + "\"");
            System.out.println("WARNING: using \"" + DefaultStrategy.class.getName() + "\", which will choose only one possible schedule");
            System.out.println("NOTE: See the \"edu.illinois.reex.scheduling.strategy\" package for a list of provided strategies\n");
            schedulingStrategy = new DefaultStrategy();
        }

        /* Set the scheduling filter to be used */
        String schedulingFilterClassName = reexProps.getProperty(ReExProperties.SCHEDULING_FILTER_KEY);
        if (schedulingFilterClassName != null) {
            try {
                schedulingFilter = (SchedulingFilter) Class.forName(schedulingFilterClassName).newInstance();
            } catch (Exception e) {
                System.err.println(UNABLE_TO_OBTAIN_INSTANCE_OF + schedulingFilterClassName + BANG);
                e.printStackTrace();
                System.exit(2);
            }
        } else {
            schedulingFilter = new DefaultFilter();
        }
    }

    /**
     * Should be called before a new exploration is going to be performed.
     * 
     * 1) Informs the scheduling strategy that is being used that a new
     * exploration will be performed. <br/>
     * 2) Prepares for a schedule execution.
     */
    public static void startingExploration(String name) {
        Listeners.fireStartingExploration(name);
        schedulingStrategy.startingExploration();
    }

    /**
     * Should be called after an exploration has been performed.
     * 
     * 1) Informs the listeners that the exploration has completed
     */
    public static void completedExploration() {
        Listeners.fireCompletedExploration();
    }

    /**
     * Should be called before a new schedule is going to be executed (during
     * exploration). Does the following:
     * 
     * 1) (Re)Initializes the scheduler state. <br/>
     * 2) Informs the scheduling strategy that is being used that a new schedule
     * will be executed.
     */
    public static void startingScheduleExecution() {
        Listeners.fireStartingSchedule();
        schedulerStateLock.lock();
        try {
            reinitState();
            schedulingStrategy.startingScheduleExecution();
            exploring = true;
        } finally {
            schedulerStateLock.unlock();
        }
    }

    /**
     * Helper method to reinitialize non-final members for each schedule.
     */
    private static void reinitState() {
        if (Listeners.debugExploration) {
            // Use ordered map when debugging to eliminate print output
            // non-determinism
            liveThreadInfos = new TreeMap<Thread, ThreadInfo>(new Comparator<Thread>() {
                @Override
                public int compare(Thread o1, Thread o2) {
                    int idComparision = ((Long) o1.getId()).compareTo(o2.getId());
                    return idComparision == 0 ? o1.getName().compareTo(o2.getName()) : idComparision;
                }
            });
        } else {
            // Use efficient map when not debugging
            liveThreadInfos = new HashMap<Thread, ThreadInfo>();
        }
        Thread currentThread = Thread.currentThread();
        liveThreadInfos.put(currentThread, new ThreadInfo(currentThread));
        pausedThreadInfos = new TreeSet<ThreadInfo>();
        blockedThreadInfos = new HashSet<ThreadInfo>();
    }

    /**
     * Should be called at the end of a schedule execution (during exploration).
     * Does the following:
     * 
     * 1) Informs the scheduling strategy that is being used that a new schedule
     * will be executed.
     */
    public static void completedScheduleExecution() {
        // TODO: schedulerStateLock.lock();
        /* Trying to make sure all threads are joined */
        while (!pausedThreadInfos.isEmpty()) {
            System.out.flush();
            System.err.println(pausedThreadInfos.size() + " thread(s) not joined! Trying to force join(s).");
            ThreadInfo joinThreadInfo = pausedThreadInfos.first();
            if (joinThreadInfo != null) {
                try {
                    performJoin(joinThreadInfo.getThread());
                } catch (InterruptedException e) {
                    // TODO Auto-generated catch block
                    e.printStackTrace();
                }
            }
        }
        // TODO: schedulerStateLock.unlock();
        /* Current schedule is done */
        exploring = false;
        Listeners.fireCompletedSchedule(schedulingStrategy.getChoicesMadeDuringThisSchedule());
        schedulingStrategy.completedScheduleExecution();
    }

    /**
     * Returns the {@link SchedulingStrategy} being used to make scheduling
     * decisions.
     * 
     * @return the {@link SchedulingStrategy} being used to make scheduling
     *         decisions.
     */
    public static SchedulingStrategy getSchedulingStrategy() {
        return schedulingStrategy;
    }

    /**
     * Returns the {@link SchedulingFilter} being used to filter scheduling
     * choices.
     * 
     * @return the {@link SchedulingFilter} being used to filter scheduling
     *         choices.
     */
    public static SchedulingFilter getSchedulingFilter() {
        return schedulingFilter;
    }

    /**
     * Returns the {@link Map} of currently live {@link Thread}s to their
     * {@link ThreadInfo}s.
     * 
     * @return the {@link Map} of currently live {@link Thread}s to their
     *         {@link ThreadInfo}s.
     */
    public static Map<Thread, ThreadInfo> getLiveThreadInfos() {
        if (Listeners.debugExploration) {
            return Collections.unmodifiableSortedMap((SortedMap<Thread, ThreadInfo>) liveThreadInfos);
        }
        return Collections.unmodifiableMap(liveThreadInfos);
    }

    /**
     * Returns the {@link Map} of currently paused {@link Thread}s to their
     * {@link ThreadInfo}s.
     * 
     * @return the {@link Map} of currently paused {@link Thread}s to their
     *         {@link ThreadInfo}s.
     */
    public static Set<ThreadInfo> getPausedThreadInfos() {
        if (Listeners.debugExploration) {
            return Collections.unmodifiableSortedSet((SortedSet<ThreadInfo>) pausedThreadInfos);
        }
        return Collections.unmodifiableSet((SortedSet<ThreadInfo>) pausedThreadInfos);
    }

    public static boolean canExecuteMoreSchedules() {
        return schedulingStrategy.canExecuteMoreSchedules();
    }

    public static void failureDetected(String errorMsg) {
        Listeners.fireFailureDetected(errorMsg, schedulingStrategy.getChoicesMadeDuringThisSchedule());
    }

    /******************************************************************
     ****************************************************************** 
     **************** SCHEDULER THREAD & CHOICE POINT *****************
     ****************************************************************** 
     ******************************************************************/

    /**
     * Start the scheduler thread!
     */
    static {
        Thread schedulerThread = new Thread(new Runnable() {
            @Override
            public void run() {
                while (true) {
                    schedulerStateLock.lock();
                    try {
                        if (!liveThreadInfos.isEmpty()) {
                            if (liveThreadInfos.size() == blockedThreadInfos.size()) {
                                failureDetected("Deadlock detected in schedule");
                                Listeners.fireCompletedExploration();
                                System.exit(2);
                            }
                            if (!pausedThreadInfos.isEmpty() && pausedThreadInfos.size() == liveThreadInfos.size() - blockedThreadInfos.size()) {
                                ThreadInfo chosenPausedThreadInfo = (ThreadInfo) choose(pausedThreadInfos, ChoiceType.THREAD_TO_SCHEDULE);
                                pausedThreadInfos.remove(chosenPausedThreadInfo);
                                chosenPausedThreadInfo.getPausingSemaphore().release();
                            }
                        }
                    } catch (Throwable exp) {
                        System.out.flush();
                        System.err.println("Uncaught exception in scheduler thread:");
                        exp.printStackTrace(System.err);
                        System.exit(2);
                    } finally {
                        schedulerWakeupCondition.awaitUninterruptibly();
                        schedulerStateLock.unlock();
                    }
                }
            }
        }, "Scheduler");
        schedulerThread.setDaemon(true);
        schedulerThread.start();
    }

    /**
     * Ask the {@link SchedulingStrategy} to choose one among the given
     * {@link Object}s. The {@link ChoiceType} specifies what is being chosen.
     * 
     * @param objectChoices
     *            {@link SortedSet} of {@link Object}s to choose from
     * 
     * @param choiceType
     *            {@link ChoiceType} indicating what is being chosen
     * 
     * @return the choice.
     */
    public static Object choose(SortedSet<? extends Object> objectChoices, ChoiceType choiceType) {
        if (objectChoices.isEmpty()) {
            throw new IllegalArgumentException("There has to be atleast one choice i.e. objectChoices cannot be empty");
        }
        schedulerStateLock.lock();
        try {
            objectChoices = schedulingFilter.filterChoices(objectChoices, choiceType);
            Object chosenObject;
            if (objectChoices.size() == 1) {
                chosenObject = objectChoices.first();
            } else {
                Listeners.fireMakingChoice(Collections.unmodifiableSortedSet(objectChoices), choiceType);
                chosenObject = schedulingStrategy.choose(Collections.unmodifiableSortedSet(objectChoices), choiceType);
                Listeners.fireChoiceMade(chosenObject);
            }
            return chosenObject;
        } finally {
            schedulerStateLock.unlock();
        }
    }

    /******************************************************************
     ****************************************************************** 
     ***************** INSTRUMENTATION HOOKS **************************
     ****************************************************************** 
     ******************************************************************/

    /**
     * Executed before a (currently executing) parent thread forks a child
     * thread (using {@link Thread}.start).
     * 
     * @param childThread
     *            the {@link Thread} that will be forked
     */
    public static void beforeForking(Thread childThread) {
        if (exploring) {
            schedulerStateLock.lock();
            try {
                ThreadInfo currentThreadInfo = liveThreadInfos.get(Thread.currentThread());
                currentThreadInfo.setEventDesc(new ThreadLifeEventDesc(EventType.FORK, childThread));
                ThreadInfo childThreadInfo = new ThreadInfo(childThread);
                liveThreadInfos.put(childThread, childThreadInfo);
                Listeners.fireBeforeForking(childThreadInfo);
            } finally {
                schedulerStateLock.unlock();
            }
        }
    }

    /**
     * Executed after a (currently executing) parent thread forks a child thread
     * (using {@link Thread}.start).
     * 
     * @param childThread
     *            the {@link Thread} that has been forked
     */
    public static void afterForking(Thread childThread) {
    }

    /**
     * Executed at the beginning of a {@link Runnable}'s run method.
     */
    public static void beginThread() {
        if (exploring) {
            schedulerStateLock.lock();
            try {
                ThreadInfo currentThreadInfo = liveThreadInfos.get(Thread.currentThread());
                currentThreadInfo.incrementRunCount();
                currentThreadInfo.setEventDesc(new ThreadLifeEventDesc(EventType.BEGIN));
            } finally {
                schedulerStateLock.unlock();
            }
        }
    }

    /**
     * Executed at the end of a {@link Runnable}'s run method.
     */
    public static void endThread() {
        if (exploring) {
            schedulerStateLock.lock();
            try {
                Thread currentThread = Thread.currentThread();
                ThreadInfo currentThreadInfo = liveThreadInfos.get(currentThread);
                currentThreadInfo.setEventDesc(new ThreadLifeEventDesc(EventType.END));
                int newRunCount = currentThreadInfo.decrementRunCount();
                if (newRunCount == 0) {
                    Set<ThreadInfo> joiningThisThreadInfos = new HashSet<ThreadInfo>();
                    for (ThreadInfo blockedThreadInfo : blockedThreadInfos) {
                        if (blockedThreadInfo.getEventDesc().getEventType().equals(EventType.JOIN)) {
                            JoinEventDesc joinEventDesc = (JoinEventDesc) blockedThreadInfo.getEventDesc();
                            if (joinEventDesc.getJoinThread().equals(currentThread)) {
                                joiningThisThreadInfos.add(blockedThreadInfo);
                                blockedThreadInfo.getPausingSemaphore().release();
                            }
                        }
                    }
                    blockedThreadInfos.removeAll(joiningThisThreadInfos);
                    liveThreadInfos.remove(currentThread);
                }
            } finally {
                schedulerWakeupCondition.signal();
                schedulerStateLock.unlock();
            }
        }
    }

    /**
     * Executed before a field is accessed.
     * 
     * @param isRead
     *            whether the access is going to be a read
     * @param owner
     *            the class to which the field belongs
     * @param name
     *            the name of the field
     * @param desc
     *            the description of the field
     */
    public static void beforeFieldAccess(boolean isRead, String owner, String name, String desc) {
        if (exploring) {
            beforeEvent(new FieldAccessEventDesc(isRead ? EventType.READ : EventType.WRITE, owner, name, desc), true);
        }
    }

    /**
     * Print an error message for unexpected usage of Unsafe class during
     * exploration
     * 
     * @param owner
     *            the class to which the method belongs
     * @param name
     *            the name of the method
     * @param desc
     *            the description of the method
     */
    public static void beforeUnsafeOther(String owner, String name, String desc) {
        if (exploring) {
            System.out.println("Reex internal error: unexpected Unsafe usage while exploring");
            System.out.println("owner=" + owner + ", name=" + name + ", desc=" + desc);
            Thread.dumpStack();
        }
    }

    /**
     * Executed after a field is accessed.
     * 
     * @param isRead
     *            whether the access is going to be a read
     * @param owner
     *            the class to which the field belongs
     * @param name
     *            the name of the field
     * @param desc
     *            the description of the field
     */
    public static void afterFieldAccess(boolean isRead, String owner, String name, String desc) {
        if (exploring) {
            afterEvent(new FieldAccessEventDesc(isRead ? EventType.READ : EventType.WRITE, owner, name, desc));
        }
    }

    /**
     * Executed instead of a monitor enter (synchronized keyword/block).
     * 
     * @param lockObject
     *            the lock/monitor to be acquired
     */
    public static void performLock(Object lockObject) {
        if (exploring) {
            performLock(lockObject, 1);
        } else {
            unsafe.monitorEnter(lockObject);
        }
    }

    /**
     * Executed instead of a monitor enter (synchronized keyword/block) and used
     * by {@link #performWait(Object)} to re-acquire released lock.
     * 
     * @param lockObject
     *            the lock/monitor to be acquired
     * @param lockCount
     *            number of instances of the lock to be acquired
     */
    private static void performLock(Object lockObject, int lockCount) {
        LockEventDesc lockEventDesc = new LockEventDesc(EventType.LOCK, lockObject);
        beforeEvent(lockEventDesc, true);
        while (true) {
            ThreadInfo lockingThreadInfo = null;
            boolean lockAvailable = true;
            schedulerStateLock.lock();
            try {
                lockingThreadInfo = liveThreadInfos.get(Thread.currentThread());
                for (ThreadInfo liveThreadInfo : liveThreadInfos.values()) {
                    if (!liveThreadInfo.equals(lockingThreadInfo) && liveThreadInfo.getLockCount(lockObject) > 0) {
                        lockAvailable = false;
                        blockedThreadInfos.add(lockingThreadInfo);
                        break;
                    }
                }
                if (lockAvailable) {
                    lockingThreadInfo.acquiredLock(lockObject, lockCount);
                    break;
                }
            } finally {
                if (!lockAvailable) {
                    schedulerWakeupCondition.signal();
                }
                schedulerStateLock.unlock();
            }
            try {
                lockingThreadInfo.getPausingSemaphore().acquire();
            } catch (InterruptedException e) {
                e.printStackTrace();
                System.exit(2);
            }
        }
        afterEvent(lockEventDesc);
    }

    /**
     * Executed instead of a monitor exit (synchronized keyword/block).
     * 
     * @param lockObject
     *            the lock/monitor to be released
     */
    public static void performUnlock(Object lockObject) {
        if (exploring) {
            LockEventDesc unlockEventDesc = new LockEventDesc(EventType.UNLOCK, lockObject);
            beforeEvent(unlockEventDesc, true);
            schedulerStateLock.lock();
            try {
                ThreadInfo unlockingThreadInfo = liveThreadInfos.get(Thread.currentThread());
                if (unlockingThreadInfo.releasedLock(lockObject, 1) <= 0) {
                    Set<ThreadInfo> waitingForThisLockThreadInfos = new HashSet<ThreadInfo>();
                    for (ThreadInfo blockedThreadInfo : blockedThreadInfos) {
                        if (blockedThreadInfo.getEventDesc().getEventType().equals(EventType.LOCK)) {
                            LockEventDesc lockEventDesc = (LockEventDesc) blockedThreadInfo.getEventDesc();
                            if (lockEventDesc.getLockObject() == lockObject) {
                                waitingForThisLockThreadInfos.add(blockedThreadInfo);
                            }
                        }
                    }
                    blockedThreadInfos.removeAll(waitingForThisLockThreadInfos);
                    pausedThreadInfos.addAll(waitingForThisLockThreadInfos);
                }
            } finally {
                schedulerStateLock.unlock();
            }
            afterEvent(unlockEventDesc);
        } else {
            unsafe.monitorExit(lockObject);
        }
    }

    /**
     * Executed instead of {@link Object#wait()} invocations.
     * 
     * @param waitObject
     *            the {@link Object} on which {@link #wait()} is being
     *            performed.
     * @throws InterruptedException
     */
    public static void performWait(Object waitObject) throws InterruptedException {
        if (exploring) {
            WaitNotifyEventDesc waitEventDesc = new WaitNotifyEventDesc(EventType.WAIT, waitObject);
            beforeEvent(waitEventDesc, true);
            ThreadInfo waitingThreadInfo = null;
            int preWaitLockCount = -1;
            schedulerStateLock.lock();
            try {
                waitingThreadInfo = liveThreadInfos.get(Thread.currentThread());
                preWaitLockCount = waitingThreadInfo.getLockCount(waitObject);
                if (preWaitLockCount < 1) {
                    throw new IllegalMonitorStateException("Calling wait without holding object lock!");
                }
                waitingThreadInfo.releasedLock(waitObject, preWaitLockCount);
                blockedThreadInfos.add(waitingThreadInfo);

                Set<ThreadInfo> waitObjectLockThreadInfos = new HashSet<ThreadInfo>();
                for (ThreadInfo blockedThreadInfo : blockedThreadInfos) {
                    if (blockedThreadInfo.getEventDesc().getEventType().equals(EventType.LOCK)) {
                        LockEventDesc lockEventDesc = (LockEventDesc) blockedThreadInfo.getEventDesc();
                        if (lockEventDesc.getLockObject() == waitObject) {
                            waitObjectLockThreadInfos.add(blockedThreadInfo);
                        }
                    }
                }
                blockedThreadInfos.removeAll(waitObjectLockThreadInfos);
                pausedThreadInfos.addAll(waitObjectLockThreadInfos);
            } finally {
                schedulerWakeupCondition.signal();
                schedulerStateLock.unlock();
            }
            waitingThreadInfo.getPausingSemaphore().acquire();
            performLock(waitObject, preWaitLockCount);
            afterEvent(waitEventDesc);
        } else {
            waitObject.wait();
        }
    }

    /**
     * Executed instead of timed {@link Object#wait()} invocations.
     * 
     * @param waitObject
     *            the {@link Object} on which {@link #wait()} is being
     *            performed.
     * @throws InterruptedException
     */
    public static void performTimedWait(Object waitObject, long millis, int nanos) throws InterruptedException {
        if (exploring) {
            WaitNotifyEventDesc waitEventDesc = new WaitNotifyEventDesc(EventType.TIMED_WAIT, waitObject);
            beforeEvent(waitEventDesc, false);
            afterEvent(waitEventDesc);
        } else {
            waitObject.wait(millis, nanos);
        }
    }

    /**
     * Executed instead of timed {@link Object#wait()} invocations.
     * 
     * @param waitObject
     *            the {@link Object} on which {@link #wait()} is being
     *            performed.
     * @throws InterruptedException
     */
    public static void performTimedWait(Object waitObject, long millis) throws InterruptedException {
        if (exploring) {
            WaitNotifyEventDesc waitEventDesc = new WaitNotifyEventDesc(EventType.TIMED_WAIT, waitObject);
            beforeEvent(waitEventDesc, false);
            afterEvent(waitEventDesc);
        } else {
            waitObject.wait(millis);
        }
    }

    /**
     * Executed instead of {@link Object#notify()} invocations.
     * 
     * @param waitObject
     *            the {@link Object} on which {@link #notify()} is being
     *            performed.
     */
    public static void performNotify(Object notifyObject) {
        if (exploring) {
            performNotify(notifyObject, false);
        } else {
            notifyObject.notify();
        }
    }

    /**
     * Executed instead of {@link Object#notifyAll()} invocations.
     * 
     * @param waitObject
     *            the {@link Object} on which {@link #notifyAll()} is being
     *            performed.
     */
    public static void performNotifyAll(Object notifyAllObject) {
        if (exploring) {
            performNotify(notifyAllObject, true);
        } else {
            notifyAllObject.notifyAll();
        }
    }

    /**
     * Executed instead of {@link Object#notify()} and
     * {@link Object#notifyAll()} invocations.
     * 
     * @param notifyObject
     *            the {@link Object} on which {@link #notify()} or
     *            {@link #notifyAll()} is being performed.
     * @param notifyAll
     *            <code>true</code> if {@link #notifyAll)}, <code>false</code>
     *            otherwise.
     */
    private static void performNotify(Object notifyObject, boolean notifyAll) {
        WaitNotifyEventDesc notifyEventDesc = new WaitNotifyEventDesc(notifyAll ? EventType.NOTIFY_ALL : EventType.NOTIFY, notifyObject);
        beforeEvent(notifyEventDesc, true);
        schedulerStateLock.lock();
        try {
            ThreadInfo notifyingThreadInfo = liveThreadInfos.get(Thread.currentThread());
            if (notifyingThreadInfo.getLockCount(notifyObject) <= 0) {
                throw new IllegalMonitorStateException("Calling notify without holding object lock!");
            }
            SortedSet<ThreadInfo> notifyableThreadInfos = new TreeSet<ThreadInfo>();
            for (ThreadInfo blockedThreadInfo : blockedThreadInfos) {
                if (blockedThreadInfo.getEventDesc().getEventType().equals(EventType.WAIT)) {
                    WaitNotifyEventDesc waitEventDesc = (WaitNotifyEventDesc) blockedThreadInfo.getEventDesc();
                    if (waitEventDesc.getWaitObject() == notifyObject) {
                        notifyableThreadInfos.add(blockedThreadInfo);
                    }
                }
            }
            if (!notifyableThreadInfos.isEmpty()) {
                if (notifyAll) {
                    blockedThreadInfos.removeAll(notifyableThreadInfos);
                    for (ThreadInfo threadToNotify : notifyableThreadInfos) {
                        threadToNotify.getPausingSemaphore().release();
                    }
                } else {
                    ThreadInfo threadToNotify = (ThreadInfo) choose(notifyableThreadInfos, ChoiceType.THREAD_TO_NOTIFY);
                    blockedThreadInfos.remove(threadToNotify);
                    threadToNotify.getPausingSemaphore().release();
                }
            }
        } finally {
            schedulerStateLock.unlock();
        }
        afterEvent(notifyEventDesc);
    }

    /**
     * Executed instead of {@link Thread#join()}.
     * 
     * @param joinThread
     *            the {@link Thread} on which {@link #join())} is being
     *            performed.
     * @throws InterruptedException
     */
    public static void performJoin(Thread joinThread) throws InterruptedException {
        if (exploring) {
            JoinEventDesc joinEventDesc = new JoinEventDesc(EventType.JOIN, joinThread);
            beforeEvent(joinEventDesc, false);
            boolean joined = false;
            schedulerStateLock.lock();
            ThreadInfo joiningThreadInfo = liveThreadInfos.get(Thread.currentThread());
            try {
                ThreadInfo joinThreadInfo = liveThreadInfos.get(joinThread);
                if (joinThreadInfo == null) {
                    joined = true;
                } else {
                    blockedThreadInfos.add(joiningThreadInfo);
                }
            } finally {
                if (!joined) {
                    schedulerWakeupCondition.signal();
                }
                schedulerStateLock.unlock();
            }
            if (!joined) {
                try {
                    joiningThreadInfo.getPausingSemaphore().acquire();
                } catch (InterruptedException e) {
                    e.printStackTrace();
                    System.exit(2);
                }
            }
            afterEvent(joinEventDesc);
        } else {
            joinThread.join();
        }
    }

    /**
     * Executed instead of a timed {@link Thread#join()}.
     * 
     * @param joinThread
     *            the {@link Thread} on which {@link #join())} is being
     *            performed.
     * @throws InterruptedException
     */
    public static void performTimedJoin(Thread joinThread, long millis, int nanos) throws InterruptedException {
        if (exploring) {
            JoinEventDesc timedJoinEventDesc = new JoinEventDesc(EventType.TIMED_JOIN, joinThread);
            beforeEvent(timedJoinEventDesc, false);
            afterEvent(timedJoinEventDesc);
        } else {
            joinThread.join(millis, nanos);
        }
    }

    /**
     * Executed instead of a timed {@link Thread#join()}.
     * 
     * @param joinThread
     *            the {@link Thread} on which {@link #join())} is being
     *            performed.
     * @throws InterruptedException
     */
    public static void performTimedJoin(Thread joinThread, long millis) throws InterruptedException {
        if (exploring) {
            JoinEventDesc timedJoinEventDesc = new JoinEventDesc(EventType.TIMED_JOIN, joinThread);
            beforeEvent(timedJoinEventDesc, false);
            afterEvent(timedJoinEventDesc);
        } else {
            joinThread.join(millis);
        }
    }

    /**
     * Executed instead of
     * {@link sun.misc.Unsafe#park(boolean isAbsolute, long time)}. Needed to
     * implement support for j.u.c classes.
     * 
     * @param isAbsolute
     *            determines milli/nano-seconds, see Unsafe
     * @param time
     *            timeout value where 0 means indefinite, see Unsafe
     */
    public static void performPark(boolean isAbsolute, long time) {
        if (exploring) {
            Thread curThread = Thread.currentThread();
            ParkUnparkEventDesc parkUnparkEventDesc = new ParkUnparkEventDesc(EventType.PARK, curThread);
            if (!isAbsolute && time == 0L) {
                beforeEvent(parkUnparkEventDesc, true);
                ThreadInfo parkingThreadInfo = null;
                boolean blocked = false;
                schedulerStateLock.lock();
                try {
                    parkingThreadInfo = liveThreadInfos.get(curThread);
                    if (parkingThreadInfo.isParkPermitAvailable()) {
                        parkingThreadInfo.setParkPermitAvailable(false);
                    } else {
                        blockedThreadInfos.add(parkingThreadInfo);
                        blocked = true;
                    }
                } finally {
                    if (blocked) {
                        schedulerWakeupCondition.signal();
                    }
                    schedulerStateLock.unlock();
                }
                if (blocked) {
                    try {
                        parkingThreadInfo.getPausingSemaphore().acquire();
                    } catch (InterruptedException e) {
                        e.printStackTrace();
                        System.exit(2);
                    }
                }
            } else {
                beforeEvent(parkUnparkEventDesc, false);
            }
            afterEvent(parkUnparkEventDesc);
        } else {
            unsafe.park(isAbsolute, time);
        }
    }

    /**
     * Executed instead of {@link sun.misc.Unsafe#unpark(Object)}. Needed to
     * implement support for j.u.c classes.
     * 
     * @param threadObject
     *            which thread to unblock, see Unsafe
     */
    public static void performUnpark(Object threadObject) {
        if (exploring) {
            ParkUnparkEventDesc parkUnparkEventDesc = new ParkUnparkEventDesc(EventType.UNPARK, (Thread) threadObject);
            beforeEvent(parkUnparkEventDesc, true);
            schedulerStateLock.lock();
            try {
                ThreadInfo unparkableThreadInfo = null;
                for (ThreadInfo blockedThreadInfo : blockedThreadInfos) {
                    if (blockedThreadInfo.getEventDesc().getEventType().equals(EventType.PARK)) {
                        ParkUnparkEventDesc parkEventDesc = (ParkUnparkEventDesc) blockedThreadInfo.getEventDesc();
                        if (parkEventDesc.getThread() == threadObject) {
                            unparkableThreadInfo = blockedThreadInfo;
                            break;
                        }
                    }
                }
                if (null != unparkableThreadInfo) {
                    blockedThreadInfos.remove(unparkableThreadInfo);
                    unparkableThreadInfo.getPausingSemaphore().release();
                } else { // need to remember for future parks
                    unparkableThreadInfo = liveThreadInfos.get((Thread) threadObject);
                    // only remember unpark if given thread has been started
                    // semantics do not guarantee anything if thread isn't
                    // started
                    if (null != unparkableThreadInfo) {
                        unparkableThreadInfo.setParkPermitAvailable(true);
                    }
                }
            } finally {
                schedulerStateLock.unlock();
            }
            afterEvent(parkUnparkEventDesc);
        } else {
            unsafe.unpark(threadObject);
        }
    }

    /**
     * Updates the {@link LocationDesc} of the current {@link Thread}'s
     * {@link ThreadInfo}. Executed before any of the events listed in
     * {@link EventType}.
     * 
     * @param className
     * @param methodName
     * @param lineNumber
     */
    public static void updateThreadLocation(String className, String methodName, int lineNumber) {
        if (exploring) {
            schedulerStateLock.lock();
            try {
                ThreadInfo currentThreadInfo = liveThreadInfos.get(Thread.currentThread());
                currentThreadInfo.setLocationDesc(new LocationDesc(className, methodName, lineNumber));
            } finally {
                schedulerStateLock.unlock();
            }
        }
    }

    /******************************************************************
     ****************************************************************** 
     ************************ HELPER METHODS **************************
     ****************************************************************** 
     ******************************************************************/

    /**
     * Helper method called before a schedule relevant event.
     * 
     * @param eventDesc
     *            {@link EventDesc} describing the schedule relevant event.
     * @param pause
     *            whether to pause the current thread before the event.
     */
    private static void beforeEvent(EventDesc eventDesc, boolean pause) {
        ThreadInfo currentThreadInfo;
        schedulerStateLock.lock();
        try {
            Listeners.fireBeforeEvent(eventDesc);
            currentThreadInfo = liveThreadInfos.get(Thread.currentThread());
            currentThreadInfo.setEventDesc(eventDesc);
            if (pause) {
                pausedThreadInfos.add(currentThreadInfo);
            }
        } finally {
            if (pause) {
                schedulerWakeupCondition.signal();
            }
            schedulerStateLock.unlock();
        }
        try {
            if (pause) {
                currentThreadInfo.getPausingSemaphore().acquire();
            }
        } catch (InterruptedException e) {
            e.printStackTrace();
            System.exit(2);
        }
    }

    /**
     * Helper method called after a schedule relevant event has been performed
     * 
     * @param eventDesc
     *            {@link EventDesc} describing the even that was performed.
     */
    private static void afterEvent(EventDesc eventDesc) {
        Listeners.fireAfterEvent(eventDesc);
    }

}
