package edu.illinois.reex.scheduling.strategy;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.PriorityQueue;
import java.util.SortedSet;

import edu.illinois.reex.ReExProperties;
import edu.illinois.reex.scheduling.ChoiceType;
import edu.illinois.reex.scheduling.ThreadInfo;
import edu.illinois.reex.scheduling.filtering.LocationFilterMode;

/**
 * Performs iterative context bounding as described in the original CHESS paper.
 * 
 * @author Vilas Jagannath <vbangal2@illinois.edu>
 * 
 */
public class CAPPStrategy extends SchedulingStrategy {

    /* Change-aware prioritization mode */
    private final LocationFilterMode mode;
    private static final String MODE_DEFAULT = "METHOD_ALL";
    private static final int CHANGED_PRIO = 0;
    private static final int UNCHANGED_PRIO = 1;

    /**
     * Work queues
     */
    private PriorityQueue<PrioritizedChoices> toExplore;

    /**
     * Current schedule choices
     */
    private List<Integer> choicesToMake;
    private List<Integer> choicesMade;

    /**
     * Preemption info
     */
    private static final int MAX_NUM_PREEMPTIONS = Integer.parseInt(ReExProperties.getInstance().getProperty(ReExProperties.PREEMPTION_BOUND_KEY,
            Integer.MAX_VALUE + ""));
    private int currentNumPreemptions;
    protected ThreadInfo previousThreadInfo;

    public CAPPStrategy() {
        /* Setting the mode property */
        String modeString = ReExProperties.getInstance().getProperty(ReExProperties.MODE_KEY, MODE_DEFAULT);
        this.mode = LocationFilterMode.valueOf(modeString);
    }

    public LocationFilterMode getMode() {
        return mode;
    }

    @Override
    public void startingExploration() {
        this.toExplore = new PriorityQueue<PrioritizedChoices>();
        this.choicesMade = new ArrayList<Integer>();
        this.choicesToMake = new ArrayList<Integer>();
    }

    @Override
    public void startingScheduleExecution() {
        if (!this.choicesMade.isEmpty()) {
            this.choicesMade.clear();
            PrioritizedChoices prioritizedChoice = this.toExplore.poll();
            this.choicesToMake = prioritizedChoice.getChoices();
        }
        this.currentNumPreemptions = 0;
        this.previousThreadInfo = null;
    }

    @Override
    public void completedScheduleExecution() {

    }

    @Override
    public boolean canExecuteMoreSchedules() {
        return !toExplore.isEmpty() || this.choicesMade.isEmpty();
    }

    @Override
    public Object choose(SortedSet<? extends Object> objectChoices, ChoiceType choiceType) {
        /*
         * Initialize choice and preemption possibility
         */
        int chosenIndex = 0;
        boolean preemptionPossible = this.previousThreadInfo != null && objectChoices.contains(this.previousThreadInfo)
                && choiceType.equals(ChoiceType.THREAD_TO_SCHEDULE);

        PriorityQueue<PrioritizedChoices> prioritizedChoices = null;

        if (this.choicesToMake.size() > this.choicesMade.size()) {
            /*
             * Make the choice to be made
             */
            chosenIndex = this.choicesToMake.get(this.choicesMade.size());
        } else {
            /*
             * Make new choices
             */

            SortedSet<? extends Object> changePrioritized = null;
            if (preemptionPossible) {
                changePrioritized = mode.filter((SortedSet<ThreadInfo>) objectChoices, this.previousThreadInfo);
            }
            prioritizedChoices = new PriorityQueue<PrioritizedChoices>();

            int choiceIndex = 0;
            for (Iterator<? extends Object> iterator = objectChoices.iterator(); iterator.hasNext(); choiceIndex++) {
                Object objectChoice = iterator.next();

                int changePriority = UNCHANGED_PRIO;
                if (!preemptionPossible || changePrioritized.contains(objectChoice)) {
                    changePriority = CHANGED_PRIO;
                }
                int preemptionPriority = this.currentNumPreemptions;
                if (preemptionPossible && !objectChoice.equals(this.previousThreadInfo)) {
                    preemptionPriority++;
                    if (preemptionPriority > MAX_NUM_PREEMPTIONS) {
                        continue;
                    }
                }
                List<Integer> priorities = Arrays.asList(new Integer[] { changePriority, preemptionPriority });

                List<Integer> choices = new ArrayList<Integer>(this.choicesMade);
                choices.add(choiceIndex);

                if (preemptionPriority <= MAX_NUM_PREEMPTIONS) {
                    prioritizedChoices.add(new PrioritizedChoices(priorities, choices));
                }
            }

            List<Integer> chosenChoices = prioritizedChoices.poll().getChoices();
            chosenIndex = chosenChoices.get(chosenChoices.size() - 1);

            this.toExplore.addAll(prioritizedChoices);
        }

        /*
         * Get chosen thread and update made choices
         */
        this.choicesMade.add(chosenIndex);
        Object chosenObject = getChosenObject(chosenIndex, objectChoices);

        /*
         * Track preemptions
         */
        if (preemptionPossible && !chosenObject.equals(this.previousThreadInfo)) {
            this.currentNumPreemptions++;
        }

        /*
         * Remember the chosen thread
         */
        if (choiceType.equals(ChoiceType.THREAD_TO_SCHEDULE)) {
            this.previousThreadInfo = (ThreadInfo) chosenObject;
        }

        /*
         * Return the chosen thread
         */
        return chosenObject;
    }

    @Override
    public List<Integer> getChoicesMadeDuringThisSchedule() {
        return this.choicesMade;
    }

}
