package edu.illinois.reex.scheduling.strategy;

import java.util.SortedSet;

import edu.illinois.reex.ReExProperties;
import edu.illinois.reex.scheduling.ChoiceType;
import edu.illinois.reex.scheduling.ThreadInfo;

/**
 * Preemption bounded search strategy (chess mode) with random choices.
 * 
 * @author Vilas Jagannath <vbangal2@illinois.edu> <br/>
 *         Qingzhou Luo <qluo2@illinois.edu>
 * 
 */
public class RandomBoundedSearchStrategy extends RandomDepthFirstStrategy {

    // Preemption bound
    private static final int MAX_NUM_PREEMPTIONS = Integer.parseInt(ReExProperties.getInstance()
            .getProperty(ReExProperties.PREEMPTION_BOUND_KEY, "2"));

    private int currentNumPreemptions;
    private ThreadInfo previousThreadInfo;

    public RandomBoundedSearchStrategy() {
        this.currentNumPreemptions = 0;
        this.previousThreadInfo = null;
    }

    @Override
    public void startingScheduleExecution() {
        super.startingScheduleExecution();
        this.currentNumPreemptions = 0;
        this.previousThreadInfo = null;
    }

    @Override
    public Object choose(SortedSet<? extends Object> objectChoices, ChoiceType choiceType) {

        /*
         * Back up the deepest choice points and call the depth first choose
         * thread.
         */
        int previousDeepestChoiceWithMoreOptions = this.deepestChoiceWithMoreOptions;
        Object chosenObject = super.choose(objectChoices, choiceType);

        /*
         * If we are not choosing a thread to schedule, or if this is the first
         * choice, there is nothing else to be done.
         */
        if (choiceType == ChoiceType.THREAD_TO_SCHEDULE && this.previousThreadInfo != null) {
            /*
             * If we are choosing the previous thread, its not preemption, so we
             * dont need to do much.
             */
            if (chosenObject.equals(this.previousThreadInfo)) {
                /*
                 * If we have reached the max number of preemptions, there are
                 * no more choices at this choice point.
                 */
                if (this.currentNumPreemptions == MAX_NUM_PREEMPTIONS) {
                    this.deepestChoiceWithMoreOptions = previousDeepestChoiceWithMoreOptions;
                }
            } else {
                /*
                 * If we are going to preempt, increment the preemption count
                 */
                int previousThreadIndex = -1;
                int index = 0;
                for (Object object : objectChoices) {
                    if (object.equals(this.previousThreadInfo)) {
                        this.currentNumPreemptions++;
                        previousThreadIndex = index;
                        break;
                    }
                    index++;
                }

                /*
                 * If we are trying to perform more preemptions than allowed,
                 * undo the preemption.
                 */
                if (currentNumPreemptions > MAX_NUM_PREEMPTIONS) {
                    chosenObject = this.previousThreadInfo;
                    this.choicesMade.set(this.choiceIndex, previousThreadIndex);
                    this.deepestChoiceWithMoreOptions = previousDeepestChoiceWithMoreOptions;
                    this.currentNumPreemptions--;
                }
            }
        }
        if (choiceType == ChoiceType.THREAD_TO_SCHEDULE) {
            this.previousThreadInfo = (ThreadInfo) chosenObject;
        }
        return chosenObject;
    }

}
