package edu.illinois.reex.scheduling.strategy;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Random;
import java.util.SortedSet;

import edu.illinois.reex.ReExProperties;
import edu.illinois.reex.scheduling.ChoiceType;

/**
 * Performs a depth first exploration of all possible schedules while making
 * random choices at choice points.
 * 
 * @author Vilas Jagannath <vbangal2@illinois.edu>
 * 
 */
public class RandomDepthFirstStrategy extends SchedulingStrategy {

    protected List<Integer> choicesMade;
    protected List<List<Integer>> choicesLeft;
    protected int choiceIndex;
    protected int deepestChoiceWithMoreOptions;
    protected boolean firstSchedule;
    protected Random random;

    public RandomDepthFirstStrategy() {
        Long seed = Long.parseLong(ReExProperties.getInstance().getProperty(ReExProperties.SEED_KEY, "" + System.currentTimeMillis()));
        random = new Random(seed);
        startingExploration();
    }

    @Override
    public void startingExploration() {
        this.choicesMade = new ArrayList<Integer>();
        this.choicesLeft = new ArrayList<List<Integer>>();
        this.firstSchedule = true;
        startingScheduleExecution();
    }

    @Override
    public void startingScheduleExecution() {
        this.choiceIndex = -1;
        this.deepestChoiceWithMoreOptions = -1;
    }

    @Override
    public void completedScheduleExecution() {
        while (this.choicesMade.size() - 1 > this.deepestChoiceWithMoreOptions) {
            this.choicesMade.remove(this.choicesMade.size() - 1);
            this.choicesLeft.remove(this.choicesLeft.size() - 1);
        }
        this.firstSchedule = false;
    }

    @Override
    public boolean canExecuteMoreSchedules() {
        return this.firstSchedule || !this.choicesMade.isEmpty();
    }

    @Override
    public Object choose(SortedSet<? extends Object> objectChoices, ChoiceType choiceType) {

        /*
         * Making next choice
         */
        this.choiceIndex++;

        /*
         * Initalize chosen index and remaining choice indices
         */
        int chosenIndex = -1;
        List<Integer> remainingChoiceIndices = null;

        /*
         * If we have encountered this choice before, make the same choice
         * unless this is the point where we need to make the next choice to
         * explore a new schedule.
         */
        if (this.choiceIndex <= this.choicesMade.size() - 1) {
            chosenIndex = this.choicesMade.get(this.choiceIndex);
            remainingChoiceIndices = this.choicesLeft.get(this.choiceIndex);
            if (this.choiceIndex == this.choicesMade.size() - 1) {
                chosenIndex = remainingChoiceIndices.remove(this.random.nextInt(remainingChoiceIndices.size()));
                this.choicesMade.set(this.choiceIndex, chosenIndex);
            }
        } else {
            chosenIndex = this.random.nextInt(objectChoices.size());
            remainingChoiceIndices = new LinkedList<Integer>();
            for (int indexChoice = 0; indexChoice < objectChoices.size(); indexChoice++) {
                if (indexChoice != chosenIndex) {
                    remainingChoiceIndices.add(indexChoice);
                }
            }
            this.choicesMade.add(chosenIndex);
            this.choicesLeft.add(remainingChoiceIndices);
        }

        /*
         * If more choices remain, mark this as the deepest choice point.
         */
        if (!remainingChoiceIndices.isEmpty()) {
            this.deepestChoiceWithMoreOptions = this.choiceIndex;
        }

        /*
         * Find the chosen object and return it
         */
        return getChosenObject(chosenIndex, objectChoices);
    }

    @Override
    public List<Integer> getChoicesMadeDuringThisSchedule() {
        return choicesMade.subList(0, choiceIndex + 1);
    }

}
