package edu.illinois.reex.scheduling.strategy;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.PriorityQueue;
import java.util.Random;
import java.util.SortedSet;

import edu.illinois.reex.ReExProperties;
import edu.illinois.reex.scheduling.ChoiceType;
import edu.illinois.reex.scheduling.ThreadInfo;

/**
 * Performs iterative context bounding as described in the original CHESS paper,
 * but with random choices.
 * 
 * @author Vilas Jagannath <vbangal2@illinois.edu>
 * 
 */
public class RandomIterativeContextBoundingStrategy extends SchedulingStrategy {

    /**
     * Random info
     */
    private final Random random;

    /**
     * Work queues
     */
    protected PriorityQueue<PrioritizedChoices> toExplore;
    protected PriorityQueue<PrioritizedChoices> toExploreNext;

    /**
     * Current schedule choices
     */
    protected List<Integer> choicesToMake;
    protected List<Integer> choicesMade;

    /**
     * Preemption info
     */
    private static final int MAX_NUM_PREEMPTIONS = Integer.parseInt(ReExProperties.getInstance().getProperty(ReExProperties.PREEMPTION_BOUND_KEY,
            Integer.MAX_VALUE + ""));
    private int currentMaxPreemptions;
    private int currentNumPreemptions;
    protected ThreadInfo previousThreadInfo;

    public RandomIterativeContextBoundingStrategy() {
        Long seed = Long.parseLong(ReExProperties.getInstance().getProperty(ReExProperties.SEED_KEY, "" + System.currentTimeMillis()));
        random = new Random(seed);
    }

    @Override
    public void startingExploration() {
        this.toExplore = new PriorityQueue<PrioritizedChoices>();
        this.toExploreNext = new PriorityQueue<PrioritizedChoices>();

        this.choicesMade = new ArrayList<Integer>();
        this.choicesToMake = new ArrayList<Integer>();

        this.currentMaxPreemptions = 0;
    }

    @Override
    public void startingScheduleExecution() {
        if (!this.choicesMade.isEmpty()) {
            this.choicesMade.clear();
            this.choicesToMake = this.toExplore.poll().getChoices();
        }
        this.currentNumPreemptions = 0;
        this.previousThreadInfo = null;
    }

    @Override
    public void completedScheduleExecution() {
        if (this.toExplore.isEmpty() && !this.toExploreNext.isEmpty()) {
            this.currentMaxPreemptions++;
            if (currentMaxPreemptions <= MAX_NUM_PREEMPTIONS) {
                this.toExplore.addAll(this.toExploreNext);
            }
            this.toExploreNext.clear();
        }
    }

    @Override
    public boolean canExecuteMoreSchedules() {
        return !(this.toExplore.isEmpty() && this.toExploreNext.isEmpty()) || this.choicesMade.isEmpty();
    }

    @Override
    public Object choose(SortedSet<? extends Object> objectChoices, ChoiceType choiceType) {
        /*
         * Initialize choice and preemption possibility
         */
        int chosenIndex = 0;
        Object chosenObject = null;
        boolean preemptionPossible = this.previousThreadInfo != null && objectChoices.contains(this.previousThreadInfo)
                && choiceType.equals(ChoiceType.THREAD_TO_SCHEDULE);

        if (this.choicesToMake.size() > this.choicesMade.size()) {
            /*
             * Make the choice to be made
             */
            chosenIndex = this.choicesToMake.get(this.choicesMade.size());
            chosenObject = getChosenObject(chosenIndex, objectChoices);
        } else {
            /*
             * Make new choices
             */
            int choiceIndex = chosenIndex;
            if (currentNumPreemptions < currentMaxPreemptions || !preemptionPossible) {
                /*
                 * We are below the preemption limit or preemption is not
                 * possible. Explore the first choice now and add the rest to
                 * toExplore.
                 */
                chosenIndex = random.nextInt(objectChoices.size());
                chosenObject = getChosenObject(chosenIndex, objectChoices);
                for (; choiceIndex < objectChoices.size(); choiceIndex++) {
                    if (choiceIndex != chosenIndex) {
                        List<Integer> choicesToExplore = new ArrayList<Integer>(this.choicesMade);
                        choicesToExplore.add(choiceIndex);
                        List<Integer> prioritiesToExplore = Arrays.asList(random.nextInt());
                        this.toExplore.add(new PrioritizedChoices(prioritiesToExplore, choicesToExplore));
                    }
                }
            } else {
                /*
                 * We have reached the preemption limit. Explore the
                 * non-preempting choice now and add the rest to toExploreNext.
                 */
                for (Iterator<? extends Object> iterator = objectChoices.iterator(); iterator.hasNext(); choiceIndex++) {
                    Object schedulableThread = iterator.next();
                    if (schedulableThread.equals(this.previousThreadInfo)) {
                        chosenIndex = choiceIndex;
                        chosenObject = schedulableThread;
                    } else {
                        List<Integer> choicesToExploreNext = new ArrayList<Integer>(this.choicesMade);
                        choicesToExploreNext.add(choiceIndex);
                        List<Integer> prioritiesToExploreNext = Arrays.asList(random.nextInt());
                        this.toExploreNext.add(new PrioritizedChoices(prioritiesToExploreNext, choicesToExploreNext));
                    }
                }
            }
        }

        /*
         * Update made choices
         */
        this.choicesMade.add(chosenIndex);

        /*
         * Track preemptions
         */
        if (preemptionPossible && !chosenObject.equals(previousThreadInfo)) {
            this.currentNumPreemptions++;
        }

        /*
         * Remember the chosen thread
         */
        if (choiceType.equals(ChoiceType.THREAD_TO_SCHEDULE)) {
            this.previousThreadInfo = (ThreadInfo) chosenObject;
        }

        /*
         * Return the chosen object
         */
        return chosenObject;
    }

    @Override
    public List<Integer> getChoicesMadeDuringThisSchedule() {
        return this.choicesMade;
    }

}
